#include "DHT.h"
#include <Arduino.h>
#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>
#include <WebSocketsServer.h>
#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <ESP8266mDNS.h>
#include <ESP8266WiFiMulti.h>
#include <WebSocketsServer.h>
#include <IotWebConf.h>

#include <OneWire.h>
#include <DallasTemperature.h>

#include <ESP8266HTTPClient.h>
#include <WiFiClient.h>

#include "pagehtml.h"

#define ON   HIGH
#define OFF  LOW

#define ROOTCELLAR 0
#define COLDROOM   1
#define OUTSIDE    2

#define NUMSENSORS 3
#define NUM_DHT22 2

#define ONE_WIRE_BUS D3 // last sensor is now a DS18B20 (outside)

#define READINTERVAL 5000

#define STRING_LEN 128

int dataPin[NUMSENSORS] = { D1, D2, D3 }; // D1, D2, D3 see https://mechatronicsblog.com/esp8266-nodemcu-pinout-for-arduino-ide/

const int LEDPIN = 2;
const char thingName[] = "RootCellarMon";
const char wifiInitialApPassword[] = "opensesame";
#define CONFIG_VERSION "wmm1"
#define CONFIG_PIN D2
#define STATUS_PIN LED_BUILTIN

// -- IotWebConf callback method declarations.
void configSaved();
boolean formValidator();

struct Record {
  int sensor;
  DHT* dht;
  float temperature;
  float humidity;
  char temperatureStr[8];
  char humidityStr[8];;
  float hiLimT;
  float loLimT;
  float hiLimH;
  float loLimH;
};

Record node[NUMSENSORS];

// Connect pin 1 (on the left) of the sensor to to 3.3V 
// Connect pin 2 of the sensor to whatever your DHTPIN is
// pin 3 is not used
// Connect pin 4 (on the right) of the sensor to GROUND
// Connect a 10K pullup resistor from pin 2 (data) to pin 1 (power) of the sensor
DHT node0( dataPin[0], DHT22); // 
DHT node1( dataPin[1], DHT22); // 

OneWire oneWire(dataPin[2]);
DallasTemperature dsSensors(&oneWire);
DeviceAddress outsideThermometer;

String smartPlugIP[2] = {
  "http://192.168.0.151/cm?cmnd=Power%20OFF",
  "http://192.168.0.153/cm?cmnd=Power%20OFF"
};
char tasmotaValue1[STRING_LEN];
char tasmotaValue2[STRING_LEN];

// TODO this should all be an object or structure
const int FAN0 = D5;
const int FAN1 = D6;
const int FAN2 = D7;
const int FAN3 = D8;

int fanState[4];
int misterState[2];

// Current LED status
bool LEDStatus;
uint32_t timeToRead;

MDNSResponder mdns;

static void writeLED(bool);
ESP8266WiFiMulti WiFiMulti;

DNSServer dnsServer;
WebServer webServer(80);
WebSocketsServer webSocket = WebSocketsServer(81);

IotWebConf iotWebConf(thingName, &dnsServer, &webServer, wifiInitialApPassword, CONFIG_VERSION);
IotWebConfParameter tasmota1 = IotWebConfParameter("Smart plug 1 IP", "tasmota1", tasmotaValue1, STRING_LEN);
IotWebConfParameter tasmota2 = IotWebConfParameter("Smart plug 2 IP", "tasmota2", tasmotaValue2, STRING_LEN);
IotWebConfSeparator separator1 = IotWebConfSeparator();
IotWebConfSeparator separator2 = IotWebConfSeparator();

unsigned long delayTime;
char payload[256];
unsigned long healthCounter;

// Commands sent through Web Socket
const char LEDON[] = "{\"command\":\"ledon\"}";
const char LEDOFF[] = "{\"command\":\"ledoff\"}";
const char RESET[] = "{\"command\":\"do_reset\"}";

const char TOGGLE0[] = "{\"command\":\"toggle0\"}";
const char TOGGLE1[] = "{\"command\":\"toggle1\"}";
const char TOGGLE2[] = "{\"command\":\"toggle2\"}";
const char TOGGLE3[] = "{\"command\":\"toggle3\"}";
const char TOGGLE4[] = "{\"command\":\"toggle4\"}";
const char TOGGLE5[] = "{\"command\":\"toggle5\"}";

const char *fanStatus[] = {"off0", "on_0", "off1", "on_1", "off2", "on_2", "off3", "on_3", "off4", "on_4"};

void webSocketEvent(uint8_t num, WStype_t type, uint8_t *payload, size_t length)
{
  // debug Serial.printf("webSocketEvent(%d, %d, ...)\r\n", num, type);
  switch(type) {
    case WStype_DISCONNECTED:
      // debug Serial.printf("[%u] Disconnected!\r\n", num);
      break;
    case WStype_CONNECTED:
      {
        IPAddress ip = webSocket.remoteIP(num);
        // debug Serial.printf("[%u] Connected from %d.%d.%d.%d url: %s\r\n", num, ip[0], ip[1], ip[2], ip[3], payload);
        // Send the current LED status
        if (LEDStatus) {
          webSocket.sendTXT(num, LEDON, strlen(LEDON));
        }
        else {
          webSocket.sendTXT(num, LEDOFF, strlen(LEDOFF));
        }
      }
      break;
    case WStype_TEXT:
      // debug Serial.printf("[%u] get Text: %s\r\n", num, payload);

      if (strcmp(LEDON, (const char *)payload) == 0) {
        writeLED(true);
      }
      else if (strcmp(LEDOFF, (const char *)payload) == 0) {
        writeLED(false);
      }
      else if (strcmp(RESET, (const char *)payload) == 0) {
      }
      else if (strcmp(TOGGLE0, (const char *)payload) == 0) {
        toggleOutput(0);
      }
      else if (strcmp(TOGGLE1, (const char *)payload) == 0) {
        toggleOutput(1);
      }
      else if (strcmp(TOGGLE2, (const char *)payload) == 0) {
        toggleOutput(2);
      }
      else if (strcmp(TOGGLE3, (const char *)payload) == 0) {
        toggleOutput(3);
      }
      else if (strcmp(TOGGLE4, (const char *)payload) == 0) {
        putSmartPlug(ROOTCELLAR);
      }
      else if (strcmp(TOGGLE5, (const char *)payload) == 0) {
        putSmartPlug(COLDROOM);
      }
      else {
        Serial.println("Unknown command");
      }
      // send data to all connected clients
      webSocket.broadcastTXT(payload, length);
      break;
    case WStype_BIN:
      // debug Serial.printf("[%u] get binary length: %u\r\n", num, length);
      hexdump(payload, length);

      // echo data back to browser
      webSocket.sendBIN(num, payload, length);
      break;
    default:
      Serial.printf("Invalid WStype [%d]\r\n", type);
      break;
  }
}

void handleRoot()
{
  // -- Let IotWebConf test and handle captive portal requests.
  if (iotWebConf.handleCaptivePortal())
  {
    // -- Captive portal request were already served.
    return;
  }
  
  size_t formFinal_len = strlen_P(INDEX_HTML) + 64; //room for form values
  char *formFinal = (char *)malloc(formFinal_len);
  if (formFinal == NULL) {
    Serial.println("formFinal malloc failed");
    return;
  }
  
  snprintf_P(formFinal, formFinal_len, INDEX_HTML);
  webServer.send(200, "text/html", formFinal);
  free(formFinal);

  unsigned int len = STRING_LEN;
  smartPlugIP[0] = String(tasmotaValue1);
  smartPlugIP[1] = String(tasmotaValue2);
  // debug Serial.print("Set URLs: "); Serial.print(smartPlugIP[0]); Serial.print(", "); Serial.println(smartPlugIP[1]); 
}

void handleNotFound()
{
  String message = "File Not Found\n\n";
  message += "URI: ";
  message += webServer.uri();
  message += "\nMethod: ";
  message += (webServer.method() == HTTP_GET)?"GET":"POST";
  message += "\nArguments: ";
  message += webServer.args();
  message += "\n";
  for (uint8_t i=0; i<webServer.args(); i++){
    message += " " + webServer.argName(i) + ": " + webServer.arg(i) + "\n";
  }
  webServer.send(404, "text/plain", message);
}

void configSaved()
{
  Serial.println("Configuration was updated.");
}

boolean formValidator()
{
  Serial.println("Validating form.");
  boolean valid = true;

  return valid;
}

static void writeLED(bool LEDon)
{
  LEDStatus = LEDon;
  // Note inverted logic for Adafruit HUZZAH board
  if (LEDon) {
    digitalWrite(LEDPIN, 0);
  }
  else {
    digitalWrite(LEDPIN, 1);
  }
}

static void toggleOutput(int f)
{
  fanState[f] = 1 - fanState[f];
  switch(f) {
    case 0:
      digitalWrite(FAN0, fanState[f]); break;
    case 1:
      digitalWrite(FAN1, fanState[f]); break;
    case 2:
      digitalWrite(FAN2, fanState[f]); break;
    case 3:
      digitalWrite(FAN3, fanState[f]); break;
    default:
      break;
  }      
}

void putSmartPlug(int smartSwitch)
{
  if ((WiFiMulti.run() == WL_CONNECTED)) {

    WiFiClient client;

    HTTPClient http;
    String commandURL = "http://";

    misterState[smartSwitch] = 1 - misterState[smartSwitch];
    if(misterState[smartSwitch] == 0){
      commandURL += smartPlugIP[smartSwitch] + "/cm?cmnd=Power%20OFF";
    }
    else{
      commandURL += smartPlugIP[smartSwitch] + "/cm?cmnd=Power%20ON";
    }

    Serial.print("Trying HTTP to: "); Serial.println(commandURL);

    if (http.begin(client, commandURL)) {  // HTTP

      // debug Serial.print("[HTTP] GET...");
      // start connection and send HTTP header
      int httpCode = http.GET();

      // httpCode will be negative on error
      if (httpCode > 0) {
        // HTTP header has been send and Server response header has been handled
        // debug Serial.printf("[HTTP] GET... code: %d\n", httpCode);

        // file found at webServer
        if (httpCode == HTTP_CODE_OK || httpCode == HTTP_CODE_MOVED_PERMANENTLY) {
          String payload = http.getString();
          // debug Serial.println(payload);
        }
      } else {
        // debug Serial.printf("[HTTP] GET... failed, error: %s\n", http.errorToString(httpCode).c_str());
      }

      http.end();
    } else {
      Serial.printf("[HTTP} Unable to connect\n");
    }
  }
}

void setup() {
  writeLED(true);
  
  timeToRead = millis();
  Serial.begin(115200);
  while(!Serial);    // time to get serial running
  Serial.println(F("Root Cellar Monitor/Controller"));
      
  pinMode(LEDPIN, OUTPUT);

  pinMode(FAN0, OUTPUT);
  pinMode(FAN1, OUTPUT);    
  pinMode(FAN2, OUTPUT);
  pinMode(FAN3, OUTPUT);
  for( int i=0; i<4; i++){
    fanState[i] = 0;
  }
  misterState[0] =0; misterState[1] = 0; 

  node[0].dht = &node0;
  node[1].dht = &node1;

  node0.begin();
  node1.begin();
  dsSensors.begin();
  if (!dsSensors.getAddress(outsideThermometer, 0)){
    Serial.println("Unable to find address for DS sensor 0"); 
  }
  else{
    Serial.print("Outside thermometer address: ");
    printAddress(outsideThermometer);
    Serial.println();
    dsSensors.setResolution(outsideThermometer, 12);
  }
  
  node[ROOTCELLAR].hiLimT = 10.0; node[ROOTCELLAR].loLimT = 2.0; node[ROOTCELLAR].hiLimH = 95.0; node[ROOTCELLAR].loLimH = 80.0; 
  node[COLDROOM].hiLimT = 14.0; node[COLDROOM].loLimT = 2.0; node[COLDROOM].hiLimH = 70.0; node[COLDROOM].loLimH = 20.0; 
  node[OUTSIDE].hiLimT = 30.0; node[OUTSIDE].loLimT = -20.0; node[OUTSIDE].hiLimH = 110.0; node[OUTSIDE].loLimH = 20.0;

  Serial.println();

  iotWebConf.setStatusPin(STATUS_PIN);
  iotWebConf.setConfigPin(CONFIG_PIN);
  iotWebConf.addParameter(&tasmota1);
  iotWebConf.addParameter(&tasmota2);
  iotWebConf.addParameter(&separator1);
  iotWebConf.setConfigSavedCallback(&configSaved);
  iotWebConf.addParameter(&separator2);
  iotWebConf.setFormValidator(&formValidator);
  iotWebConf.getApTimeoutParameter()->visible = true;
  iotWebConf.init();

  webServer.on("/", handleRoot);
  webServer.on("/config", []{ iotWebConf.handleConfig(); });
  webServer.onNotFound([](){ iotWebConf.handleNotFound(); });

  writeLED(false);

  webServer.begin();
  webSocket.begin();
  webSocket.onEvent(webSocketEvent);
  
  delay(2000);
}

// function to print a device address
void printAddress(DeviceAddress deviceAddress)
{
  for (uint8_t i = 0; i < 8; i++)
  {
    if (deviceAddress[i] < 16) Serial.print("0");
    Serial.print(deviceAddress[i], HEX);
  }
}

void controlLogic(){
  float jitter = 1.0;
  float deltaT = 2.0;
  /*
  node[ROOTCELLAR].hiLimT = 10.0; node[ROOTCELLAR].loLimT = 2.0; node[ROOTCELLAR].hiLimH = 95.0; node[ROOTCELLAR].loLimH = 80.0; 
  node[COLDROOM].hiLimT = 14.0; node[COLDROOM].loLimT = 2.0; node[COLDROOM].hiLimH = 70.0; node[COLDROOM].loLimH = 20.0; 
  node[OUTSIDE].hiLimT = 30.0; node[OUTSIDE].loLimT = -20.0; node[OUTSIDE].hiLimH = 110.0; node[OUTSIDE].loLimH = 20.0;
   */
  // if the temperature is too high in the cold room and it is colder outside, turn on the cold room inlet fan
  // if the temperature is too low in the cold room, send an alarm
  // if the temperature is too high in the root cellar and it is colder outside
  //    and the humidity is not too low outside turn on the inlet fan
  if(node[ROOTCELLAR].temperature > node[ROOTCELLAR].hiLimT + jitter
    && node[ROOTCELLAR].temperature - node[OUTSIDE].temperature > deltaT){
      //digitalWrite(FAN0, 1);
      Serial.println("Root cellar inlet on");
    }
  // if the temperature is too low in the cold room, send an alarm
  // if the humidity is too low in root cellar, turn on the humidifier
  
}

void loop() { 
  uint32_t now;
  webSocket.loop();
  webServer.handleClient();
  healthCounter=millis();
  now = millis();

  iotWebConf.doLoop();
  
  if(timeToRead < now){
    writeLED(true);
      
    // debug Serial.println("Reading...");
    timeToRead = now + READINTERVAL; // Note: this will still work across a rollover

    for(int i=0; i<NUM_DHT22; i++){
      float x;
      strcpy(node[i].temperatureStr, "Rerr");
      strcpy(node[i].humidityStr, "Rerr");
      x = node[i].dht->readHumidity();
      if(!isnan(x)){ 
        node[i].humidity = x; 
        dtostrf(node[i].humidity, 3, 1, node[i].humidityStr);
      }    
      x = node[i].dht->readTemperature();
      if(!isnan(x)){ 
        node[i].temperature = x;
        dtostrf(node[i].temperature, 3, 1, node[i].temperatureStr);
      }
    }

    // debug Serial.println("Reading DS18B20");
    dsSensors.requestTemperatures(); // Send the command to get temperatures
    node[2].temperature = dsSensors.getTempC(outsideThermometer);
    dtostrf(node[2].temperature, 3, 1, node[2].temperatureStr);
    strcpy(node[2].humidityStr, "N/A");
    
    controlLogic();
    writeLED(false);
  }

  if(healthCounter % 1000 == 0){
    os_sprintf(payload, 
      "{\"heartbeat\":%ld, \"tRoot\":\"%sC\",\"hRoot\":\"%s%%\",\"tCold\":\"%sC\",\"hCold\":\"%s%%\",\"tOut\":\"%sC\",\"hOut\":\"%s%%\",\"m0\":%d,\"m1\":%d,\"m2\":%d,\"m3\":%d,\"humid1\":%d,\"humid2\":%d}", 
      healthCounter/1000, 
      node[0].temperatureStr, node[0].humidityStr, 
      node[1].temperatureStr, node[1].humidityStr,
      node[2].temperatureStr, node[2].humidityStr,
      fanState[0], fanState[1], fanState[2], fanState[3],
      misterState[0], misterState[1]);

    // debug Serial.println(payload);
    webSocket.broadcastTXT(payload, strlen(payload));
  }
}
